@@ -1,3 +1,4 @@
+from math import radians, cos, sin, asin, sqrt
 from typing import Union
 
 import numpy as np
@@ -121,3 +122,34 @@ def is_point_within_line_projection(line: Union[np.ndarray, list], point: Union[
     b = (0 <= np.inner(v, s) <= np.inner(s, s))
 
     return b
+
+
+def haversine(lon1: float, lat1: float, lon2: float, lat2: float):
+    """ Calculate the great circle distance in kilometers between two points on the earth (specified in decimal degrees)
+        Source: https://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points
+    Parameters
+    ----------
+    lon1 : float
+        Longitude of first point
+    lat1 : float
+        Latitude of first point
+    lon2 : float
+        Longitude of second point
+    lat2 : float
+        Latitude of second point
+    Returns
+    -------
+    float
+        Distance between the points in meters
+
+    """
+    # convert decimal degrees to radians
+    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
+
+    # haversine formula
+    dlon = lon2 - lon1
+    dlat = lat2 - lat1
+    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
+    c = 2 * asin(sqrt(a))
+    r = 6371000  # Radius of earth in meters
+    return c * r
